import os
import sys
import copy
from unittest import TestCase
import unittest
import io
import pickle
import rbi_tree.tree
from rbi_tree.tree import ITree, ITreed
import gc

class CustomClass(object):
    def __init__(self, a, b):
        self.a = a
        self.b = b

class TestTreeCase(TestCase):
    def test_simple(self):
        for value in ['abc', 100, {0:1}, list('abc')]:
             t = ITree()
             i1 = t.insert(60, 80)
             i2 = t.insert(20, 40, value=value)

             ivs = t.find(10, 30)
             self.assertEqual(len(ivs), 1)
             iv = ivs[0]
             self.assertEqual(iv[2], value)

    def test_iterate(self):
        t = ITree()
        for pos in reversed(range(10)):
            num = t.insert(pos, pos+1)

        starts, ends, ids = list(zip(*t.iter_ivl()))
        self.assertEqual(starts, tuple(range(10)))
        self.assertEqual(ends, tuple([i+1 for i in range(10)]))


    def test_leak(self):
        trees = []
        for tren in range(10):
            t = ITree()
            for pos in range(100):
                t.insert(pos, pos+1, CustomClass(tren+1, pos))
            trees.append(t)
            gc.collect()
        for t in trees:
            for s,e,v in t.iter_ivl():
                self.assertGreater(v.a, 0)
                

    def test_pickling(self):
        # Test for bug in https://github.com/mikpom/rbi_tree/issues/4
        t = ITree()
        for pos in range(100):
            num = t.insert(pos, pos+1)

        s = io.BytesIO()
        pickle.dump(t, s)

        for i in range(100):
            t3 = pickle.loads(s.getvalue())
            del t3
            gc.collect()

        t3 = pickle.loads(s.getvalue())
        self.assertEqual(len(list(t3.iter_ivl())), 100)
        
    def test_copy(self):
        t = ITree()
        for pos in range(10):
            num = t.insert(pos, pos+1)

        t2 = copy.copy(t)
        starts,ends,ids = list(zip(*t2.iter_ivl()))
        self.assertEqual(starts, tuple(range(10)))
        self.assertEqual(ends, tuple([i+1 for i in range(10)]))

    def test_find_at(self):
        t = ITree()
        for i, int in enumerate([(1,3),(5,9),(6,10)]):
            t.insert(*int, i)
        res = t.find_at(2)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], (1,3,0))

class TestTreedCase(TestCase):
    def test_copy(self):
        t = ITreed()
        for pos in range(10):
            num = t.insert(pos, pos+1)

        t2 = copy.copy(t)
        starts,ends,ids = list(zip(*t2.iter_ivl()))
        self.assertEqual(starts, tuple(range(10)))
        self.assertEqual(ends, tuple([i+1 for i in range(10)]))
    
    def test_tree_with_deletion(self):
        t = ITreed()
        i1 = t.insert(60,80)
        i2 = t.insert(20,40)

        # Test find
        ivs = t.find(10,30)
        self.assertEqual(len(ivs), 1)
        self.assertEqual(ivs[0], (20, 40, i2))
        self.assertEqual(t.get_ivl(i2), [20, 40])

        # Test back-to-back non-overlapping
        self.assertEqual(len(t.find(40,41)), 0)
        self.assertEqual(len(t.find(19,20)), 0)
        # But finds at point
        ivs = t.find_at(20)
        self.assertEqual(len(ivs),1)

        # Test remove nonexistent
        with self.assertRaises(ValueError):
            t.remove(1000)

        # Test remove nonexistent
        with self.assertRaises(TypeError):
            t.remove('abc')
            
        # Test remove
        ivl = t.find(60, 80)
        self.assertEqual(len(ivl), 1)

        s,e,sid = ivl[0]
        t.remove(sid)
        ivs = t.find(50,70)
        self.assertEqual(len(ivs),0)

        
if __name__=='__main__':
    unittest.main()




# Copyright 2020-2021 Mikhail Pomaznoy
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

# distutils: language = c++
# distutils: extra_compile_args = -std=c++11
from cython.operator cimport dereference as deref, preincrement as inc, address
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp.utility cimport pair
from cpython.ref cimport PyObject, Py_INCREF, Py_DECREF

cdef extern from "intervaltree.hpp" namespace "Intervals":
    cdef cppclass Interval[T1,T2]:
        Interval(T1 a, T1 b)
        Interval(T1 a, T1 b, T2 val)
        T1 high
        T1 low
        T2 value
    cdef cppclass IntervalTree[T1,T2]:
        IntervalTree()
        void clear()
        IntervalTree[T1,T2] IntervalTree(IntervalTree[T1,T2] other)
        bint insert(Interval&& interval)
        void findOverlappingIntervals(Interval iterval, vector[Interval] out)
        void findIntervalsContainPoint(int point, vector[Interval] out)
        vector[Interval[T1,T2]] intervals()
        IntervalTree[T1,T2] IntervalTree(vector[Interval[T1,T2]] ivls)
        bint remove(Interval interval)
        

ctypedef Interval[int, int] CIntervalInt
ctypedef IntervalTree[int, int] CTreeInt
ctypedef map[int, CIntervalInt*] Ivlmap
ctypedef pair[int, CIntervalInt*] keyval

ctypedef Interval[int, PyObject*] CIntervalObj
ctypedef IntervalTree[int, PyObject*] CTreeObj

cdef class ITree:
    cdef CTreeObj* tree

    def __cinit__(self):
        self.tree = new CTreeObj()

    def __dealloc__(self):
        cdef vector[CIntervalObj] intervals = self.tree.intervals()
        cdef vector[CIntervalObj].iterator it = intervals.begin()
        
        while it != intervals.end():
            val  = <object>(deref(it).value)
            if not val is None:
                Py_DECREF(val)
            inc(it)

        del self.tree

    def __reduce__(self):
        intervals = list(self.iter_ivl())
        return (ITree._from_intervals, (intervals,))

    def insert(self, start, end, value=None):
        """Insert an interval [start, end) """
        
        cdef CIntervalObj* ivl = new CIntervalObj(
            start, end, <PyObject*>value)
        self.tree.insert(deref(ivl))
        val = <object>value
        if not val is None:
            Py_INCREF(val)
        del ivl
        return

    def find(self, int start, int end):
        """Search intervals overlapping [start, end). Returns list of 
        overlapping intervals' ids."""
        cdef CIntervalObj* ivl = new CIntervalObj(start,end)
        cdef vector[CIntervalObj] out
        self.tree.findOverlappingIntervals(deref(ivl), out)
        del ivl
        a = []
        cdef vector[CIntervalObj].iterator it = out.begin()
        while it != out.end():
            # Have to exclude for the sake of half-openness
            if deref(it).high!=start and deref(it).low!=end:
                val = <object>deref(it).value
                a.append( (deref(it).low, deref(it).high, val) )
            inc(it)
        return a

    def _from_intervals(cls, intervals=None):
        tree = cls()
        if not intervals is None:
            for start, end, val in intervals:
                tree.insert(start, end, val)
        return tree
    _from_intervals = classmethod(_from_intervals)

    def find_at(self, int point):
        """Search for intervals containing specified point. Returns list of 
        overlapping intervals' ids."""
        cdef vector[CIntervalObj] out
        self.tree.findIntervalsContainPoint(point, out)
        a = []
        cdef vector[CIntervalObj].iterator it = out.begin()
        while it != out.end():
            if not deref(it).high == point:
                val = <object>deref(it).value
                a.append( (deref(it).low, deref(it).high, val) )
            inc(it)
        return a

    def iter_ivl(self):
        """Iterate over all intervals. Yields tuples (start, end, id)."""
        cdef vector[CIntervalObj] intervals = self.tree.intervals()
        cdef vector[CIntervalObj].iterator it = intervals.begin()
        while it != intervals.end():
            val = <object>deref(it).value
            yield (deref(it).low, deref(it).high, val)
            inc(it)

cdef class ITreed():
    """Version of ITree supporting deletion
    """
    cdef CTreeInt* tree
    cdef Ivlmap ivldata
    cdef map[int, CIntervalInt*].const_iterator datapos
    cdef tot

    def __cinit__(self):
        self.tree = new CTreeInt()
        self.ivldata = Ivlmap()
        self.datapos = self.ivldata.begin()
        self.tot = 0

    def __init__(self, tot=None):
        if not tot is None:
            self.tot = tot

    def __dealloc__(self):
        del self.tree

    def __reduce__(self):
        intervals = list(self.iter_ivl())
        return (ITreed._from_intervals, (intervals, self.tot))

    def _from_intervals(cls, intervals=None, tot=0):
        tree = cls(tot)
        if not intervals is None:
            for start, end, val in intervals:
                tree._insert(start, end, val)

        return tree
    _from_intervals = classmethod(_from_intervals)


    def _insert(self, start, end, value):
        cdef CIntervalInt* ivl = new CIntervalInt(start, end, value)
        self.tree.insert(deref(ivl))
        self.datapos = self.ivldata.insert(self.datapos,
                            keyval(ivl.value, ivl))
        
    def insert(self, start, end):
        """Insert an interval [start, end) and returns an id
        of the interval. Ids are incrementing integers, i.e. 0,1,2 etc."""
        cdef int ivl_id = self.tot
        self._insert(start, end, ivl_id)
        self.tot += 1
        return ivl_id

    cdef CIntervalInt* _get_interval(self, id):
        cdef map[int, CIntervalInt*].iterator it = self.ivldata.find(id)
        if it != self.ivldata.end():
            return deref(it).second
        else:
            return NULL
        
    def find(self, int start, int end):
        """Search intervals overlapping [start, end). Returns list of 
        overlapping intervals' ids."""
        cdef CIntervalInt* ivl = new CIntervalInt(start,end)
        cdef vector[CIntervalInt] out
        self.tree.findOverlappingIntervals(deref(ivl), out)
        del ivl
        a = []
        cdef vector[CIntervalInt].iterator it = out.begin()
        while it != out.end():
            # Have to exclude for the sake of half-openness
            if deref(it).high!=start and deref(it).low!=end:
                a.append( (deref(it).low, deref(it).high, deref(it).value) )
            inc(it)
        return a

    def get_ivl(self, id):
        """Return a list [start,end] of the interval with specified id."""
        cdef CIntervalInt* ivl = self._get_interval(id)
        return [deref(ivl).low, deref(ivl).high]

    def find_at(self, int point):
        """Search for intervals containing specified point. Returns list of 
        overlapping intervals' ids."""
        cdef vector[CIntervalInt] out
        self.tree.findIntervalsContainPoint(point, out)
        a = []
        cdef vector[CIntervalInt].iterator it = out.begin()
        while it != out.end():
            if not deref(it).high == point:
                a.append( (deref(it).low, deref(it).high, deref(it).value) )
            inc(it)
        return a

    def remove(self, int id):
        """Delete interval with specified id."""
        cdef CIntervalInt* ivl = self._get_interval(id)
        if not ivl is NULL:
            self.tree.remove(deref(ivl))
        else:
            raise ValueError
        self.ivldata.erase(id)
        del ivl

    def iter_ivl(self):
        """Iterate over all intervals. Yields tuples (start, end, id)."""
        cdef vector[CIntervalInt] intervals = self.tree.intervals()
        cdef vector[CIntervalInt].iterator it = intervals.begin()
        while it != intervals.end():
            yield (deref(it).low, deref(it).high, deref(it).value)
            inc(it)



from setuptools import setup, Extension
from pkg_resources import resource_filename as pkg_file

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

file_ext = '.pyx' if USE_CYTHON else '.cpp'

ext = Extension(
    'rbi_tree.tree',
    sources=[pkg_file('rbi_tree', 'tree'+file_ext)],
)

if USE_CYTHON:
    setup(ext_modules=cythonize(ext))
else:
    setup(ext_modules=[ext])


